/*
 *
 * exSGBD.cpp:
 *
 *         Implements G.B. Dantzig's Aircraft Allocation model,
 *         in: G.B. Dantzig, Linear Programming and Extensions
 *         (Princeton U. Press, 1963) pp. 572-597.
 *
 *         See also: A.J. King, "Stochastic programming problems:
 *         examples from the literature",
 *          in: Yu. Ermoliev and R.J-B Wets, eds, Numerical Techniques for
 *         Stochastic Optimization (Springer-Verlag, 1988) pp. 543-567.
 *         (compile with /DOSLMSDLL on INTEL platforms.)
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "scn_c_api.h"

#define INF 1.0e31            /*  plus infinity */

int main( int argc, char *argv[] ) {


    /* Model dimensions */
    int ncol=27, nrow=9, nels=44;

    /* Sparse matrix data...organized by row */
    int mrow[]={ 1, 1, 1, 1, 1,
            2, 2, 2, 2,
            3, 3, 3,
            4, 4, 4, 4, 4,
            5, 5, 5, 5,
            6, 6, 6, 6, 6, 6,
            7, 7, 7, 7, 7,
            8, 8, 8, 8, 8, 8,
            9, 9, 9, 9, 9, 9 };
    int mcol[]={ 1, 2, 3, 4, 5,
            6, 7, 8, 9,
            10, 11, 12,
            13, 14, 15, 16, 17,
            1,        13, 18, 19,
            2, 6, 10, 14, 20, 21,
            3, 7,     15, 22, 23,
            4, 8, 11, 16, 24, 25,
            5, 9, 12, 17, 26, 27 };
    double dels[] = { 1.0, 1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0, 1.0,
                16.0,              9.0, -1.0, 1.0,
                15.0, 10.0,  5.0, 11.0, -1.0, 1.0,
                28.0, 14.0,       22.0, -1.0, 1.0,
                23.0, 15.0,  7.0, 17.0, -1.0, 1.0,
                81.0, 57.0, 29.0, 55.0, -1.0, 1.0 };

    /* Objective */
    double dobj[]={ 18.0, 21.0, 18.0, 16.0, 10.0, 15.0, 16.0, 14.0, 9.0,
              10.0,  9.0,  6.0, 17.0, 16.0, 17.0, 15.0, 10.0, 0.0,
              13.0,  0.0, 13.0,  0.0,  7.0,  0.0,  7.0,  0.0, 1.0 };

    /* Column bounds */
    double dclo[]={ 0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
              0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
              0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0 };
    double dcup[]={ INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,
              INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,
              INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF };

    /* Row bounds */
    double drlo[]={ -INF, -INF, -INF, -INF,  0.0, 0.0, 0.0, 0.0, 0.0 };
    double drup[]={ 10.0, 19.0, 25.0, 15.0,  0.0, 0.0, 0.0, 0.0, 0.0 };

    /* Stages */
    int nstg=2;
    int n_first_stg_rows=4;
    int n_first_stg_cols=17;
    int rstg[]={ 1,1,1,1,2,2,2,2,2 };
    int cstg[]={ 1,1,1,1,1,1,1,1,1,
            1,1,1,1,1,1,1,1,2,
            2,2,2,2,2,2,2,2,2 };

    /* Stochastic data */
    int nindp=5;
    int nsamp[]={ 5, 2, 5, 5, 3 };
    double demand[]={ 200, 220, 250, 270, 300,
                50, 150,
                140, 160, 180, 200, 220,
                10, 50, 80, 100, 340,
                580, 600, 620 };
    double dprobs[]={ 0.2, 0.05, 0.35, 0.2, 0.2,
                0.3, 0.7,
                0.1, 0.2, 0.4, 0.2, 0.1,
                0.2, 0.2, 0.3, 0.2, 0.1,
                0.1, 0.8, 0.1 };

    /* local variables */
    int rc,ns=1,ii,iii,jj,*indx,*incr;
    double dd,dp=1.0;


    EKKContext *env=ekks_initializeContext();

    for (ii=0;ii<nindp;ii++) ns *= nsamp[ii];     /* Compute number of scenarios */

    EKKStoch *stoch=ekks_newStoch(env,"SGBDS",ns); /* Create an EKKStoch object */


    rc=ekks_createCore(stoch,nstg,rstg,cstg,nrow,ncol,nels,dobj,drlo,
                       drup,dclo,dcup,mrow,mcol,dels,0 ); /* Create Core */

    /* Generate Scenarios */


    memset( dobj,0,ncol*sizeof(double) );         /* ...zero everything */
    memset( drlo,0,nrow*sizeof(double) );
    memset( drup,0,nrow*sizeof(double) );
    memset( dclo,0,ncol*sizeof(double) );
    memset( dcup,0,ncol*sizeof(double) );
    nels = 0;

    /* ...initialization -- indx points to first sample of each rv */
    indx = (int *) malloc( (1+nindp)*sizeof(int) );
    memset( indx,0,(1+nindp)*sizeof(int));
    for (jj=0;jj<nindp;jj++) {
     indx[jj+1] += indx[jj] + nsamp[jj];
     dp *= dprobs[ indx[jj] ];
     drlo[ n_first_stg_rows + jj ] = demand[ indx[jj] ];
     drup[ n_first_stg_rows + jj ] = demand[ indx[jj] ];
    }


    rc=ekks_addScenario(stoch,1,0,2,dp,1,
          nrow,ncol,nels,dobj,drlo,drup,dclo,dcup,
          mrow,mcol,dels,ADD_TO_CORE_VALUES );    /* ...add first sample */

    /* ...initialize the sample space increment for each rv to 1 */
    incr = (int *) malloc( nindp*sizeof(int) );
    for (jj=0;jj<nindp;jj++) incr[jj] = 1;

    /***** ...main loop
       The logic of this loop is as follows:
        While the sample size nsamp[jj] divides the scenario index ii,
          reverse the increment direction incr[jj],
           and increase the random variable index jj by 1.
        Increment the jj'th random variable by incr[jj]
          and call ekks_addScenario with the new sample.
    ***** */

    for (ii=1;ii<ns;ii++) {
     iii=ii; jj=0;
     while ( !(iii%nsamp[jj]) ) {
         iii /= nsamp[jj];
         incr[jj] = -incr[jj];
         jj++;
     }
     dd = dprobs[ indx[jj] ];
     indx[jj] += incr[jj];
     dp = dp* dprobs[ indx[jj] ]/dd;
     drlo[ n_first_stg_rows + jj ] = demand[ indx[jj] ];
     drup[ n_first_stg_rows + jj ] = demand[ indx[jj] ];
     rc=ekks_addScenario(stoch,ii+1,0,2,dp,1,
               nrow,ncol,nels,dobj,drlo,drup,dclo,dcup,     /* Add the next scenario */
               mrow,mcol,dels,ADD_TO_CORE_VALUES );
    }


/* Describe Full model and don't generate deterministic equivalent */
	ekks_describeFullModel(stoch,0);  

/* Set decomposition solver parameters */
    ekks_setMaxsubmodels(stoch,10);                      
	ekks_setScaleOn(stoch);
	ekks_setCrash(stoch,3);
	ekks_setPresolve(stoch,3);
	ekks_setFinalSimplexSolverOff(stoch);

/* Solve using nested decomposition with 
	minimum number of rows for each subproblem */
	ekks_setMinNumRows(stoch,750);
    ekkse_nestedLSolve(stoch,CUTBYROWSIZE,0);  
    
/* Check prod_mix optimal value */
	assert(fabs(ekks_getRobjvalue(stoch)-1566.042)<1.0e-6*1566.042);

    ekks_deleteStoch(stoch);                      /* Delete EKKStoch object */
    ekks_endContext(env);                         /* Uninitialize OSL environment */

    printf("\n ************************************");
    printf("\n ***   All testing successful !   ***");
    printf("\n ************************************\n");fflush(stdout);
    
    return 0;
}
