/*
 *
 * exSGBDt.cpp:
 *
 *         Input distribution is sampled and processed by
 *         tree* subroutines.
 *
 *		   Problem is formulated as Stochastic Quadratic Program
 *         and solved with OSL's QP barrier method.
 *
 *         Implements G.B. Dantzig's Aircraft Allocation model,
 *         in: G.B. Dantzig, Linear Programming and Extensions
 *         (Princeton U. Press, 1963) pp. 572-597.
 *
 *         See also: A.J. King, "Stochastic programming problems:
 *         examples from the literature",
 *          in: Yu. Ermoliev and R.J-B Wets, eds, Numerical Techniques for
 *         Stochastic Optimization (Springer-Verlag, 1988) pp. 543-567.
 *         (compile with  /DOSLMSDLL on INTEL platforms.)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "scn_c_api.h"

#define INF 1.0e31          /* plus infinity */

#define N_SAMPLES  300      /* number of samples to be generated */

int main( int argc, char *argv[] ) {

    /* Model dimensions */
    int ncol=27, nrow=9, nels=44;

    /* Sparse matrix data...organized by row */
    int mrow[]={ 1, 1, 1, 1, 1,
            2, 2, 2, 2,
            3, 3, 3,
            4, 4, 4, 4, 4,
            5, 5, 5, 5,
            6, 6, 6, 6, 6, 6,
            7, 7, 7, 7, 7,
            8, 8, 8, 8, 8, 8,
            9, 9, 9, 9, 9, 9 };
    int mcol[]={ 1, 2, 3, 4, 5,
            6, 7, 8, 9,
            10, 11, 12,
            13, 14, 15, 16, 17,
            1,        13, 18, 19,
            2, 6, 10, 14, 20, 21,
            3, 7,     15, 22, 23,
            4, 8, 11, 16, 24, 25,
            5, 9, 12, 17, 26, 27 };
    double dels[] = { 1.0, 1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0, 1.0,
                16.0,              9.0, -1.0, 1.0,
                15.0, 10.0,  5.0, 11.0, -1.0, 1.0,
                28.0, 14.0,       22.0, -1.0, 1.0,
                23.0, 15.0,  7.0, 17.0, -1.0, 1.0,
                81.0, 57.0, 29.0, 55.0, -1.0, 1.0 };

    /* Objective */
    double dobj[]={ 18.0, 21.0, 18.0, 16.0, 10.0, 15.0, 16.0, 14.0, 9.0,
              10.0,  9.0,  6.0, 17.0, 16.0, 17.0, 15.0, 10.0, 0.0,
              13.0,  0.0, 13.0,  0.0,  7.0,  0.0,  7.0,  0.0, 1.0 };
    /* Diagonal entries */
    double dqdg[]={ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
              0.0,  0.0,  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
              13.0,  0.0, 13.0,  0.0,  7.0,  0.0,  7.0,  0.0, 1.0 };


    /* Column bounds */
    double dclo[]={ 0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
              0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
              0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0 };
    double dcup[]={ INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,
              INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,
              INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF,  INF };

    /* Row bounds */
    double drlo[]={ -INF, -INF, -INF, -INF,  0.0, 0.0, 0.0, 0.0, 0.0 };
    double drup[]={ 10.0, 19.0, 25.0, 15.0,  0.0, 0.0, 0.0, 0.0, 0.0 };

    /* Stages */
    int nstg=2;
    int n_first_stg_rows=4;
    int n_first_stg_cols=17;
    int rstg[]={ 1,1,1,1,2,2,2,2,2 };
    int cstg[]={ 1,1,1,1,1,1,1,1,1,
            1,1,1,1,1,1,1,1,2,
            2,2,2,2,2,2,2,2,2 };

    /* Stochastic data */
    int nindp=5;
    int nsamp[]={ 5, 2, 5, 5, 3 };
    double demand[]={ 200, 220, 250, 270, 300,
                50, 150,
                140, 160, 180, 200, 220,
                10, 50, 80, 100, 340,
                580, 600, 620 };
    double dprobs[]={ 0.2, 0.05, 0.35, 0.2, 0.2,
                0.3, 0.7,
                0.1, 0.2, 0.4, 0.2, 0.1,
                0.2, 0.2, 0.3, 0.2, 0.1,
                0.1, 0.8, 0.1 };


    /* local variables */
    int rc,ns=1,jj,*indx;
    double dd,dp=1.0,dwgt;
    int irand,indx_beg,ev_indx;
    int nscen,ianc,istg;
    int *label;
    int evoffset;


    EKKContext *env=ekks_initializeContext();            /*  Initialize OSL environments */

    EKKStoch *stoch=ekks_newStoch(env,"SGBDT",N_SAMPLES); /* create an EKKStoch object */

    rc=ekks_setCoreProblemType(stoch, quadratic);        /* set  core problem as quadratic */
    if (rc>1) { printf ("\n rc=%d\n", rc); return(rc); }

    rc=ekks_createCore(stoch,nstg,rstg,cstg,nrow,ncol,nels,dobj,drlo,drup,
                       dclo,dcup,mrow,mcol,dels,dqdg ); /* Create Core */
    if ( rc>1 ) return(rc);


    /* ...zero everything */
    memset( dobj,0,ncol*sizeof(double) );
    memset( drlo,0,nrow*sizeof(double) );
    memset( drup,0,nrow*sizeof(double) );
    memset( dclo,0,ncol*sizeof(double) );
    memset( dcup,0,ncol*sizeof(double) );
    nels = 0;

    /* ...malloc event indexing array */
    indx = (int *) malloc( (1+nindp)*sizeof(int) );
    memset( indx,0,(1+nindp)*sizeof(int) );

    /* ...malloc label array */
    label = (int *) malloc( (nstg-1)*sizeof(int) );


    /* Generate Scenarios from sampling */

    for (ns=0; ns<N_SAMPLES; ns++) {

    /*   ...for each indep rv, process random number to get index */
        indx_beg = 0;
        for (jj=0;jj<nindp;jj++) {
                ev_indx = 0;
                irand = rand();
                dd = (double)irand/(double)RAND_MAX;
                dp = dprobs[indx_beg+ev_indx];
                while ( irand > dp*RAND_MAX )
                        dp += dprobs[++ev_indx + indx_beg];
                indx[jj] = ev_indx;
                indx_beg += nsamp[jj];
        }

    /*   ...process indep events into a unique "event label"
        (there should be one such label for each stage in the simulation
          -- this example has only one stage ) */

        label[0]= indx[0];
        for (jj=1;jj<nindp;jj++)
                label[0] = label[0]*nsamp[jj] + indx[jj];

        dp = 1.0e0;
        dp /= N_SAMPLES;

    /*   ...initialize tree if necessary, or add event to tree */
        if (ns==0) {
                rc=ekks_createScenarioTree(stoch,label,nstg-1,N_SAMPLES,dp );
                if (rc != 0 ) return(1);
        } else {
                rc=ekks_addEventToTree(stoch,label,nstg-1,dp);
                if (rc<0) { printf ("\n rc=%d\n", rc); return(-rc); }
        }

    /*   ...get scenario branching information from tree */
        rc=ekks_getScenarioFromTree(stoch,ns,&ianc,&istg,&dwgt,&nscen,&label );
        if (rc>0) return(rc);

        if (istg == 1) ianc = 0;       /* always true in this example */

        istg++;                        /* always need to do this */

    /*   ...assign values to right-hand sides based on index */
        evoffset=0;
        for (jj=0;jj<nindp;jj++) {
                drlo[ n_first_stg_rows + jj ] = demand[ indx[jj]+evoffset ];
                drup[ n_first_stg_rows + jj ] = demand[ indx[jj]+evoffset ];
                evoffset += nsamp[jj];
        }

    /* ...add sample */
        rc=ekks_addScenario(stoch,nscen,ianc,istg,dwgt,1,
                            nrow,ncol,nels,dobj,drlo,drup,dclo,dcup,
                            mrow,mcol,dels,ADD_TO_CORE_VALUES );
        if (rc>1) return(rc);
    }


    
/* Describe Full model and populate with data */
	rc=ekks_describeFullModel(stoch,1);                     
    assert(rc >0);

/* Solve with OSL quadratic barrier method */
    EKKModel *ekkm = ekkse_getCurrentModel(stoch);	
    rc = ekk_quadraticBarrier(ekkm,0);

/* Print Objective Distribution */
	ekks_printObjectiveDistribution(stoch);

/* Delete all objects */
	ekks_deleteTree(stoch);
    ekks_deleteStoch(stoch);   
    ekks_endContext(env);      

    printf("\n ************************************");
    printf("\n ***   All testing successful !   ***");
    printf("\n ************************************\n");fflush(stdout);
    
    return 0;
}
